

#import libraries
print("importing packages")

import re
import molgrid
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.nn import init
from torch import autograd
from scipy.stats import pearsonr
import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
import wandb
import pandas as pd


config_defaults = {
        'init_channel_size': 16,
        'models' : "scalar",
        'epochs': 1,
        'lr': 0.001,
        'dropout': 0,
        'momentum': 0.9,
        'solver': "adam",
        'batch_norm':0,
        'weight_decay':0,
        'binary_rep':False,
        'batch_size':32,
        'step_size': 100
    }
wandb.login()
wandb.init(config = config_defaults)
config = wandb.config



#Define the path
print("Defining the path")

path_train_lig_cache = '/net/pulsar/home/koes/liyeny/dl_project/molcache_file/lig_train0.molcache2'
path_train_rec_cache ='/net/pulsar/home/koes/liyeny/dl_project/molcache_file/rec_train0.molcache2'
path_train_type = '/net/pulsar/home/koes/liyeny/dl_project/type_files/gaffwmintrain0.types'

path_test_lig_cache = '/net/pulsar/home/koes/liyeny/dl_project/molcache_file/lig_test0.molcache2'
path_test_rec_cache = '/net/pulsar/home/koes/liyeny/dl_project/molcache_file/rec_test0.molcache2'
path_test_type = '/net/pulsar/home/koes/liyeny/dl_project/type_files/gaffwmintest0.types'

#Organize data into examples   
print("Organizing data into examples ")    

traine = molgrid.ExampleProvider(ligmolcache=path_train_lig_cache, recmolcache=path_train_rec_cache, shuffle=True, default_batch_size=config.batch_size, iteration_scheme = molgrid.IterationScheme.LargeEpoch, duplicate_first=False, balanced=True) ##a pair of ligand protein
traine.populate(path_train_type) ##provide the affinity

teste = molgrid.ExampleProvider(ligmolcache=path_test_lig_cache, recmolcache=path_test_rec_cache, shuffle=True, default_batch_size=1, iteration_scheme=molgrid.IterationScheme.LargeEpoch, duplicate_first=False, balanced=True)#balanced: number bad = good 
teste.populate(path_test_type)

gmaker = molgrid.GridMaker(binary=config.binary_rep) #binary ture-- point
dims = gmaker.grid_dimensions(14*2) # 14 atoms -- 1 protein 1 ligand # only one rec+onelig per example
tensor_shape = (config.batch_size,)+dims


#define, create, and initialize the model
print("defining, creating, and initializing the model ")  
    

def weights_init(m):
    if isinstance(m, nn.Conv3d) or isinstance(m, nn.Linear):
        init.xavier_uniform_(m.weight.data)
        if m.bias is not None:
            init.constant_(m.bias.data, 0)

if config.models == "scalar":
    from equi_scalar_model import Net
    model = Net(dims, config.init_channel_size)
    
elif config.models == "vector_A":
    from equi_vector_model import Net_A
    model = Net_A(dims, config.init_channel_size)
    
elif config.models == "vector_B":
    from equi_vector_model import Net_B
    model = Net_B(dims, config.init_channel_size)
    
elif config.models == "vector_C":
    from equi_vector_model import Net_C
    model = Net_C(dims, config.init_channel_size)
    
elif config.models == "vector_D":
    from equi_vector_model import Net_D
    model = Net_D(dims, config.init_channel_size)
    
elif config.models == "vector_E":
    from equi_vector_model import Net_E
    model = Net_E(dims, config.init_channel_size)
    
elif config.models == "tensor_A":
    from equi_tensor_model import Net_A
    model = Net_A(dims, config.init_channel_size)
    
elif config.models == "tensor_B":
    from equi_tensor_model import Net_B
    model = Net_B(dims, config.init_channel_size)
    
elif config.models == "tensor_C":
    from equi_tensor_model import Net_C
    model = Net_C(dims, config.init_channel_size)
    
elif config.models == "tensor_D":
    from equi_tensor_model import Net_D
    model = Net_D(dims, config.init_channel_size)
    
elif config.models == "tensor_E":
    from equi_tensor_model import Net_E
    model = Net_E(dims, config.init_channel_size)
    
elif config.models == "tensor_F":
    from equi_tensor_model import Net_F
    model = Net_F(dims, config.init_channel_size)
    
    
elif config.models == "tensor_G":
    from equi_tensor_model import Net_G
    model = Net_G(dims, config.init_channel_size)
    
elif config.models == "tensor_H":
    from equi_tensor_model import Net_H
    model = Net_H(dims, config.init_channel_size)  
      
elif config.models == "default":
    from default2018_single_model import Net
    model = Net(dims, config.init_channel_size)
#model = Net(dims, config.init_channel_size)



if torch.cuda.device_count() > 1:
        print("Using {} GPUs".format(torch.cuda.device_count()))
        model = nn.DataParallel(model)
else:
        print('GPUS: {}'.format(torch.cuda.device_count()), flush=True)
        
model.to('cuda:0')
model.apply(weights_init)




#define custom loss and sigmoid
print("defining custom loss")

def Aff_loss(outputs, labels):
    
    outputs = outputs.flatten()
    labels = labels.flatten()
    
    loss = torch.tensor((0.0), device='cuda')
    count = 0
    
    for output, label in zip(outputs, labels):
        if label < 0:
            abs_label = - label
    
            if output> abs_label:   
                loss += (output-abs_label)**2
                count +=1
        else:
            loss += (output-label)**2  
            count +=1
    return loss/count


def Pose_loss(outputs, labels):
    outputs = outputs.flatten()
    labels = labels.flatten()
    
    lossfunc = nn.BCEWithLogitsLoss()
    loss = lossfunc(outputs, labels)
    
    return loss


def Sigmoid(z):
    return 1/(1 + np.exp(-z))    





#define optimizer and scheduler
print("defining optimizer and scheduler")

optimizer = optim.SGD(model.parameters(), lr=config.lr, weight_decay=config.weight_decay)
if config.solver == "adam":
    optimizer = optim.Adam(model.parameters(), lr=config.lr, weight_decay=config.weight_decay)
    
scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=config.step_size, gamma=0.1) ###




#define zero tensors to put things in
print("defining zero tensors to put things in")

#for training
input_tensor = torch.zeros(tensor_shape, dtype=torch.float32, device='cuda') # input data
affinity_labels = torch.zeros(config.batch_size, dtype=torch.float32, device ='cuda')
pose_labels = torch.zeros(config.batch_size, dtype=torch.float32, device ='cuda')
#for testing
singleinput = torch.zeros((1,)+dims, dtype=torch.float32, device='cuda') #input data
singleAffinity_label = torch.zeros(1,dtype=torch.float32,device='cuda')
singlePose_label = torch.zeros(1,dtype=torch.float32,device='cuda')


def Plot(label, predict,val=True):
    
    df = pd.DataFrame({'label': label,'predict': predict})
    scatter = sns.jointplot(x = "label",y = "predict", data = df)
    plt.xlabel("label")
    plt.ylabel("predict")
    #wandb.init(project="test", config= config)
    plot = wandb.Image(plt)
    if val == True:
        wandb.log({"val/plot": plot})
    else:
        wandb.log({"train/plot": plot})



def train_single_fold(model):
   
    wandb.run.name = config.models + "/" + str(config.init_channel_size)
    
    #losses = [] #losses for training for each batch 
    #mses = [] #mean square error of test for each epoch during training 
    #ames = [] #absolute mean error of test for each epoch during training
    #pearsons = [] #correlation of test set(all samples) for each epoch during training
    
    #record the final predictions
    train_aff_labels = []
    train_aff_outputs = []

    for e in range(config.epochs): 
        for idx, batch in enumerate(traine):
            optimizer.zero_grad()
            batch.extract_label(1, affinity_labels)#get labels for affinity
            aff_labels = affinity_labels.float().to('cuda')
            batch.extract_label(0, pose_labels) #get labels for pose
            p_labels = pose_labels.float().to('cuda')
            gmaker.forward(batch, input_tensor, random_translation=2, random_rotation=True)
            aff_output, p_output = model(input_tensor)
            aff_loss = Aff_loss(aff_output,aff_labels)
            pose_loss = Pose_loss(p_output, p_labels)
            loss = 0.1* aff_loss + pose_loss
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(),1)
            optimizer.step()
    
            
            if e == config.epochs -1:  #record the labels and predictions of the training set in the final epoch
                aff_output = aff_output.squeeze().tolist()
                aff_labels = aff_labels.squeeze().tolist()
                train_aff_outputs = train_aff_outputs + aff_output
                train_aff_labels = train_aff_labels + aff_labels
                
        wandb.log({"train/total_loss":loss,"epoch": e})
        wandb.log({"train/aff_loss":aff_loss,"epoch": e})
        wandb.log({"train/pose_loss":pose_loss,"epoch": e})
        if e == config.epochs -1:    
            Plot(train_aff_labels, train_aff_outputs,False)
        
        ### calculate accuracy for pose classification
        #p_output = np.array(p_output.detach().cpu().numpy()).flatten()
        #p_labels = np.array(p_labels.detach().cpu().numpy()).flatten()
        
        #p_probs = Sigmoid(p_output)  # the probabilities
        #p_classes = (p_probs > np.array(([0.5])) ).astype(np.float32) #the classes 
        #accuracy = (p_classes == p_labels).astype(np.float).sum()/len(p_labels)
        #wandb.log({"train/accuracy":accuracy, "epoch": e})

        scheduler.step()
        
        
        
        #eval performance on test set
        with torch.no_grad():
            val_aff_labels = []
            val_aff_outputs = []
            val_p_labels = []
            val_p_outputs = []
            
            for idx, batch in enumerate(teste):
                
                gmaker.forward(batch,singleinput)
                val_aff_output, val_p_output = model(singleinput)
                #pose
                batch.extract_label(0, singlePose_label)
                val_p_labels.append(singlePose_label.detach().cpu().numpy())
                val_p_outputs.append(val_p_output.detach().cpu().numpy()) #####
                
                #affinity
                batch.extract_label(1,singleAffinity_label)
                if singleAffinity_label < 0:
                    continue  
                val_aff_outputs.append(val_aff_output.detach().cpu().numpy())
                val_aff_labels.append(singleAffinity_label.detach().cpu().numpy()) 
            
            #flatten things
            val_p_outputs = np.array(val_p_outputs).flatten()
            val_p_labels = np.array(val_p_labels).flatten()
            val_aff_outputs = np.array(val_aff_outputs).flatten()
            val_aff_labels = np.array(val_aff_labels).flatten()
            
            ### calculate accuracy for pose classification
            val_p_probs = Sigmoid(val_p_outputs)  # the probabilities
            val_p_classes = (val_p_probs > np.array(([0.5])) ).astype(np.float32) #the classes 
            accuracy = (val_p_classes == val_p_labels).astype(np.float).sum()/len(val_p_labels)
            wandb.log({"val/accuracy":accuracy, "epoch": e})
            
            #### performance for affinity
            valrmse = np.sqrt(np.mean((val_aff_outputs - val_aff_labels)**2))  
            if np.isinf(valrmse):
                valrmse = 1000
                
            #valame = np.mean(np.abs(results-testlabels))
            R = pearsonr(val_aff_outputs,val_aff_labels)[0]
            #mses.append(valrmse)
            #ames.append(valame)
            #pearsons.append(R)
            wandb.log({"val/rmses":valrmse, "epoch": e})
            #wandb.log({"val/ames":valame, "epoch": e})
            wandb.log({"val/pearsons":R, "epoch": e})
            if e == config.epochs -1: #record the labels/predictions of the validation in the final epoch
                Plot(val_aff_labels,val_aff_outputs,True)

     

    return model



trainedModel = train_single_fold(model)







